@@ -1,8 +1,7 @@
-import gym
 import numpy as np
 
 
-class TicTacToe(gym.Env):
+class TicTacToe:
     """
     This class handles all the logic of running a tic tac
     toe game. It expects as input two players which themselves
@@ -37,7 +36,7 @@ class TicTacToe(gym.Env):
             self.board[8] = 1
             self.current_player = 1
 
-        #run_test()
+        run_test()
 
         self.board_string = """
                 | {s1} | {s2} | {s3} |    
@@ -104,7 +103,7 @@ class TicTacToe(gym.Env):
 
         self.current_player = 1
 
-        #run_test()
+        run_test()
 
     def render(self, mode="human"):
         board_dict = {f"s{i}": int(self.board[i - 1]) for i in range(1, 10)}
