@@ -9,6 +9,7 @@ from collections import defaultdict
 from operator import itemgetter
 
 import numpy as np
+import time
 
 
 class MCTS:
@@ -227,13 +228,14 @@ class MINIMAX:
             res.append((child,v))
             continue
 
-
         max_c = max(res,key=itemgetter(1))
 
         all_maxes = [child[0] for child in res if child[1] == max_c[1]]
 
         return np.random.choice(all_maxes)
 
+        #return all_maxes
+
     def alpha_beta_depth(self,node,cutoff_test=None,n=5):
 
         player = node.player
@@ -297,6 +299,14 @@ class MINIMAX:
 
             max_child = self.alpha_beta(node)
 
+            """
+            actions = []
+            for child in max_child:
+                actions.append(self.get_parent_action(node,child))
+
+            return actions
+            """
+
         elif type == "alphabeta_depth":
 
             if node in self.children:
@@ -435,7 +445,6 @@ class ZeroMCTS:
         self._env = env
         self._nn = nn
         self._Qsa = defaultdict(float)
-        self._Ns = defaultdict(int)
         self._Nsa = defaultdict(int)
         self._Ps = defaultdict(float)
         self._Vs = defaultdict(float)
@@ -474,7 +483,6 @@ class ZeroMCTS:
         leaf_player = leaf.player
 
         for node, a in reversed(path):
-            self._Ns[node] += 1
             if node == leaf:
                 continue
 
@@ -489,16 +497,31 @@ class ZeroMCTS:
             else:
                self._Qsa[(node,a)] += r
 
+    def calculate_NS(self,node):
+        valid_actions = self._env.valid_actions(node.state)
+        return sum([self._Nsa[(node,a)] for a in valid_actions])
+
     def display_state_info(self,node):
         """
         More of a debugging method
         TICTACTOE ONLY
         """
-        node_index = self.dec_pts.index(node)
+        if node not in self._Vs or node not in self._Ps:
+            p, v = self._nn.predict(self._state_encoder(node))
+
+            self._Ps[node] = p
+            self._Vs[node] = v
+
+        try:
+            node_index = self.dec_pts.index(node)
+        except:
+            print("oppsssy")
+
+        action_size = self._env.action_size
         value = self._Vs[node]
-        counts = self._Ns[node]
+        counts = self.calculate_NS(node)
         valid_actions = self._env.valid_actions(node.state)
-        blank_board = np.zeros((9,))
+        blank_board = np.zeros((action_size,))
         q_board = blank_board.copy()
         v_board = blank_board.copy()
         n_board = blank_board.copy()
@@ -512,17 +535,17 @@ class ZeroMCTS:
         v_board[child_actions] = child_v
         n_board[valid_actions] = child_n
 
-        prior_p = self._Ps[node].copy().reshape((3,3))
+        prior_p = self._Ps[node].copy()
         print("Vs")
         print(value)
         print("Ns")
         print(counts)
         print("Nsa")
-        print(n_board.reshape((3,3)))
+        print(n_board)
         print("Qsa")
-        print(q_board.reshape((3,3)))
+        print(q_board)
         print("Vsa")
-        print(v_board.reshape((3,3)))
+        print(v_board)
         print("Prior P")
         print(prior_p)
 
@@ -554,7 +577,11 @@ class ZeroMCTS:
         @return:
         """
 
-        node_index = self.dec_pts.index(node)
+        try:
+            node_index = self.dec_pts.index(node)
+        except:
+            sim_nodes = self.get_nodes_from_state(node.state)
+            print("ahhh")
 
         if node.terminal():
             return True
@@ -581,6 +608,17 @@ class ZeroMCTS:
             children.append(next_node)
         return children
 
+    def get_nodes_from_state(self,state):
+        """
+        Return all nodes of tree that have the
+        given state
+        """
+        nodes = []
+        for node in self.dec_pts:
+            if np.equal(node.state,state).all():
+                nodes.append(node)
+        return nodes
+
     def policy(self,node,tau,max=False):
         """
         Give the current statistics in the tree
@@ -624,7 +662,6 @@ class ZeroMCTS:
     def reset_tree(self):
 
         self._Qsa = defaultdict(float)
-        self._Ns = defaultdict(int)
         self._Nsa = defaultdict(int)
         self._Ps = defaultdict(float)
         self._Vs = defaultdict(float)
@@ -647,6 +684,22 @@ class ZeroMCTS:
             node_index = self.dec_pts.index(node)
             node = self.dec_pts[node_index] # Perfer to use the node that already exists
 
+        if node not in self.dec_pts:
+            print("WHYU ")
+            #self.add_dec_pt(no)
+            if node != self.root:
+                parent_index = self.dec_pts.index(node.parent)
+                self.children[parent_index].append(node)
+                a = node.parent_action
+                self._Qsa[(node.parent, a)] = 0
+                self._Nsa[(node.parent, a)] = 0
+
+            self.dec_pts.append(node)
+            self.children.append([])
+
+            print("Wtf")
+
+
         path = self.search(node)
 
         leaf,a = path[-1]
@@ -665,7 +718,6 @@ class ZeroMCTS:
         while True:
             path.append(node)
             if node not in self.dec_pts:
-                self._Ns[node] = 0 # just init hear will increment during backup
                 return path
 
             node_index = self.dec_pts.index(node)
@@ -725,6 +777,8 @@ class ZeroMCTS:
         else:
             p, v = self._nn.predict(self._state_encoder(node))
 
+            assert type(p) == np.ndarray
+
             self._Ps[node] = p
 
             self._Vs[node] = v
@@ -736,7 +790,7 @@ class ZeroMCTS:
 
         assert p.sum() > 0.9999
 
-        n_s = self._Ns[node]
+        n_s = self.calculate_NS(node)
 
         if node == self.root and not self.act_max:
 
@@ -767,17 +821,17 @@ class ZeroMCTS:
             r = node.reward()
             return r
 
+        """
         if node in self._Vs:
             r = self._Vs[node]
 
             return r
+        """
+        p,v = self._nn.predict(self._state_encoder(node))
 
-        else:
-            p,v = self._nn.predict(self._state_encoder(node))
+        self._Vs[node] = v
 
-            self._Vs[node] = v
-
-            return v
+        return v
 
     @property
     def points(self):
